#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/29 16:12
# @Author  : Wu_RH
# @FileName: image_create.py
from ..abs.board import AbstractBoard
from ..config.config import IMAGE_CONFIG, DEFAULT_CONFIG
from ..utils.element_renderer import *


def _hex_to_rgb(hex_color: str):
    hex_color = hex_color.lstrip('#')
    return tuple(int(hex_color[i:i + 2], 16) for i in (0, 2, 4))


def get_text(
    text: str,
    width: float = "auto",
    height: float = "auto",
    cover_pos_label: bool = True,
    color: tuple[str, str] = ("#FFFFFF", "#000000"),
    dominant_by_height: bool = True,
    style: str = "",
):
    """
    :param text:文本内容
    :param width: 宽度
    :param height: 高度
    :param cover_pos_label: 覆盖格子内的X=N标识
    :param dominant_by_height: 高主导的对齐 否则宽主导
    :param color: 色号字符串#RRGGBB 第一个表示黑底 第二个表示白底 '#FFFFFF'表示白色
    :param style: (web) 样式内容
    """
    if dominant_by_height is None:
        dominant = None
    else:
        dominant = "height" if dominant_by_height else "width"
    return {
        "type": "text",
        "text": text,
        "content": text,
        'color_black': _hex_to_rgb(color[0]),
        'color_white': _hex_to_rgb(color[1]),
        'width': width,
        'height': height,
        "font_size": 1,
        "cover": cover_pos_label,
        "dominant": dominant,
        "style": style,
    }


def get_image(
    image_path: str,
    image_width: float = "auto",
    image_height: float = "auto",
    cover_pos_label: bool = True,
    dominant_by_height: bool = True,
    style: str = "",
):
    """
    :param image_path:图片在data下的路径位置
    :param image_width:图片的水平缩放比例
    :param image_height:图片的垂直缩放比例
    :param cover_pos_label:是否覆盖X=N标识
    :param dominant_by_height: 高主导的对齐 否则宽主导
    :param style: (web) 样式内容
    """
    if dominant_by_height is None:
        dominant = None
    else:
        dominant = "height" if dominant_by_height else "width"
    return {
        'type': 'image',
        'image': image_path,  # 图片对象
        'height': image_height,  # 高度（单元格单位或auto）
        'width': image_width,   # 宽度（单元格单位或auto）
        "cover": cover_pos_label,
        "dominant": dominant,
        "style": style,
    }


def get_row(
    *args,
    spacing=0,
    dominant_by_height=True
):
    """
    水平排列元素
    :param args: 子元素列表
    :param spacing: 每个元素之间的间距值
    :param dominant_by_height: 高主导的对齐 否则宽主导
    """
    if dominant_by_height is None:
        dominant = None
    else:
        dominant = "height" if dominant_by_height else "width"
    for child in args:
        if child["dominant"] is None:
            child["dominant"] = "height"
    height = [e["height"] for e in args if type(e["height"]) is int]
    if height:
        height = max(height)
    else:
        height = "auto"
    return {
        "type": "row",
        "children": args,
        "spacing": spacing,
        "cover": all(e["cover"] for e in args),
        "height": height,
        "width": "auto",
        "dominant": dominant
    }


def get_col(
    *args,
    spacing=0,
    dominant_by_height=False
):
    """
    水平排列元素
    :param args: 子元素列表
    :param spacing: 每个元素之间的间距值
    :param dominant_by_height: 高主导的对齐 否则宽主导
    """
    if dominant_by_height is None:
        dominant = None
    else:
        dominant = "height" if dominant_by_height else "width"
    for child in args:
        if child["dominant"] is None:
            child["dominant"] = "width"
    width = [e["width"] for e in args if type(e["width"]) is int]
    if width:
        width = max(width)
    else:
        width = "auto"
    return {
        "type": "col",
        "children": args,
        "spacing": spacing,
        "cover": all(e["cover"] for e in args),
        "height": "auto",
        "width": width,
        "dominant": dominant
    }


def get_dummy(
        width: float = 0.01,
        height: float = 0.01
) -> object:
    """
    创建占位符元素

    :param width: 宽度（单元格单位）
    :param height: 高度（单元格单位）
    """
    return {
        "type": "placeholder",
        "width": width,
        "height": height,
        "cover": True,
        "dominant": None
    }


def draw_board(
        board: AbstractBoard,
        background_white: bool = False,
        bottom_text: str = "",
        cell_size: int = 100,
        output="output"
) -> bytes:
    """
    绘制多个题板图像，支持横向拼接。
    :param board: AbstractBoard 实例，支持 get_board_keys。
    :param background_white: 是否白底。
    :param bottom_text: 底部文字。
    :param cell_size: 单元格大小。
    :param output: 输出文件名（不含扩展名）。
    """
    def load_font(size: int) -> ImageFont.FreeTypeFont:
        path = pathlib.Path(minesweepervariants.__path__[0])
        path /= CONFIG["assets"]
        path /= CONFIG["font"]["name"]
        try:
            return ImageFont.truetype(path, size)
        except OSError:
            return ImageFont.load_default()

    def cfg(_path):
        return _hex_to_rgb(CONFIG[_path]["white_bg" if background_white else "black_bg"])

    def int_to_roman(num: int) -> str:
        # 定义数值与罗马符号的映射表（按数值降序排列）
        val_symbols = [
            (1000, "M"),
            (900, "CM"),
            (500, "D"),
            (400, "CD"),
            (100, "C"),
            (90, "XC"),
            (50, "L"),
            (40, "XL"),
            (10, "X"),
            (9, "IX"),
            (5, "V"),
            (4, "IV"),
            (1, "I")
        ]

        _roman = []  # 存储结果字符
        for _value, symbol in val_symbols:
            # 当剩余数字大于等于当前值
            while num >= _value:
                num -= _value  # 减去当前值
                _roman.append(symbol)  # 添加对应符号
            if num == 0:  # 提前终止
                break
        return "".join(_roman)

    board_keys = board.get_board_keys()
    configs = {k: {
        "by_mini": board.get_config(k, "by_mini"),
        "pos_label": board.get_config(k, "pos_label"),
        "row_col": board.get_config(k, "row_col"),
    } for k in board_keys}

    CONFIG = {}

    CONFIG.update(IMAGE_CONFIG)
    CONFIG["output_path"] = DEFAULT_CONFIG["output_path"]

    margin_ratio = CONFIG["margin"]["top_left_right_ratio"]
    bottom_ratio = CONFIG["margin"]["bottom_ratio"]
    axis_ratio = CONFIG["axis_label"]["font_ratio"]
    mini_ratio = CONFIG["corner"]["mini_font_ratio"]

    # 色彩配置
    bg_color = _hex_to_rgb(CONFIG["background"]["white" if background_white else "black"])
    text_color = _hex_to_rgb(CONFIG["text"]["white" if background_white else "black"])
    grid_color = cfg("grid_line")
    dye_color = cfg("dye")
    stroke_color = cfg("stroke")
    pos_label_color = cfg("pos_label")

    margin = cell_size * margin_ratio
    bottom_margin = cell_size * bottom_ratio

    sizes = {}
    for key in board_keys:
        br = board.boundary(key=key)
        rows = len(board.get_row_pos(br))
        cols = len(board.get_col_pos(br))
        sizes[key] = (cols, rows)

    total_width = int(sum(c for _, c in sizes.values()) * cell_size + (len(board_keys) + 1) * margin)
    max_rows = max(r for r, _ in sizes.values())
    total_height = int(margin + max_rows * cell_size + bottom_margin)

    image = Image.new("RGBA", (total_width, total_height), bg_color)
    draw = ImageDraw.Draw(image)

    x_offset = margin
    for key in board_keys:
        rows, cols = sizes[key]
        by_mini = configs[key]["by_mini"]
        pos_label = configs[key]["pos_label"]
        row_col = configs[key]["row_col"]

        # 题板左上角编号
        if len(board_keys) > 2:
            roman = int_to_roman(board_keys.index(key) + 1)

            roman_margin = margin * 1.3
            max_w = roman_margin  # 目标最大宽度不应超过 margin 区域
            low, high = 8, int(roman_margin * 0.63)
            best = low
            while low <= high:
                mid = (low + high) // 2
                font = load_font(mid)
                if font.getlength(roman) <= max_w:
                    best = mid
                    low = mid + 1
                else:
                    high = mid - 1

            font = load_font(best)

            center_x = x_offset - cell_size * 0.3

            paste_y = int(margin * 0.5)
            draw.text((center_x, paste_y),
                      roman,
                      fill=text_color,
                      font=font,
                      anchor="mm",
                      stroke_width=1 if best <= 14 else 0,
                      stroke_fill=stroke_color)

        # 坐标轴标签
        if row_col:
            axis_font_size = int(cell_size * axis_ratio)
            axis_font = load_font(axis_font_size)

            for col in range(cols):
                x = x_offset + col * cell_size + cell_size // 2
                y = margin / 2
                text = chr(64 + col // 26) if col > 25 else ''
                text += chr(65 + col % 26)
                draw.text((x, y), text, fill=text_color, font=axis_font, anchor="mm")

            for row in range(rows):
                x = x_offset - cell_size * 0.25
                y = margin + row * cell_size + cell_size / 2
                draw.text((x, y), str(row + 1), fill=text_color, font=axis_font, anchor="mm")

        # 染色
        for pos, _ in board(key=key):
            r, c = pos.x, pos.y
            x0 = x_offset + c * cell_size
            y0 = margin + r * cell_size
            x1, y1 = x0 + cell_size, y0 + cell_size
            if board.get_dyed(pos):
                draw.rectangle([x0, y0, x1, y1], fill=dye_color)

        # 网格线
        line_width = CONFIG["grid_line"]["width"]
        for r in range(rows + 1):
            y = margin + r * cell_size
            draw.line([(int(x_offset - cell_size * (line_width * 0.3)), y),
                       (int(x_offset + cols * cell_size + cell_size * (line_width * 0.3)), y)],
                      fill=grid_color, width=int(cell_size * line_width))
        for c in range(cols + 1):
            x = x_offset + c * cell_size
            draw.line([(x, int(margin - cell_size * (line_width * 0.3))),
                       (x, int(margin + rows * cell_size + cell_size * (line_width * 0.3)))],
                      fill=grid_color, width=int(cell_size * line_width))

        # X=N标签
        if pos_label:
            label_font = load_font(int(cell_size * CONFIG["pos_label"]["size"]))
            for pos, obj in board(mode="object", key=key):
                if board.get_type(pos) == "C":
                    continue
                r, c = pos.x, pos.y
                x = x_offset + c * cell_size + cell_size / 2
                y = margin + r * cell_size + cell_size / 2
                txt = chr(64 + c // 26) if c > 25 else ''
                txt += chr(65 + c % 26)
                txt += f"={r}"
                draw.text((x, y), txt, fill=pos_label_color, font=label_font, anchor="mm")

        # 内容渲染 - 使用ElementRenderer
        for pos, obj in board(mode="object", key=key):
            r, c = pos.x, pos.y
            x0_cell = x_offset + c * cell_size
            y0_cell = margin + r * cell_size
            value = board.get_value(pos)
            if value is None:
                continue

            # 创建元素渲染器
            renderer = Renderer(
                cell_size=cell_size,
                background_white=background_white,
                origin=(x0_cell, y0_cell),
                font_path=CONFIG["font"]["name"],
                assets=CONFIG["assets"]
            )

            renderer.render(image, value.compose(board))

        # 渲染角标
        if by_mini:
            for pos, obj in board(mode="object", key=key):
                r, c = pos.x, pos.y
                x0_cell = x_offset + c * cell_size
                y0_cell = margin + r * cell_size
                x1_cell = x0_cell + cell_size
                y1_cell = y0_cell + cell_size
                value = board.get_value(pos)
                if value is None:
                    continue

                if type(obj) is type(board.get_config(key, "VALUE")):
                    continue
                if type(obj) is type(board.get_config(key, "MINES")):
                    continue

                mini_font = load_font(int(cell_size * mini_ratio))
                draw.text((x1_cell - cell_size * 0.02, y1_cell + cell_size * 0.05),
                          value.type().decode('utf-8', 'ignore'),
                          fill=text_color, font=mini_font, anchor="rd")

        x_offset += cols * cell_size + margin

    # 底部文本
    if bottom_text:
        bottom_y = margin + max_rows * cell_size
        max_w = total_width
        low, high = 8, int(bottom_margin * 0.63)
        best = low
        while low <= high:
            mid = (low + high) // 2
            font = load_font(mid)
            if font.getlength(bottom_text) <= max_w:
                best = mid
                low = mid + 1
            else:
                high = mid - 1

        font = load_font(best)
        h = font.getbbox(bottom_text)[3] - font.getbbox(bottom_text)[1]
        y = bottom_y + (bottom_margin / 2) + (h / 4)
        y = min(y, total_height - h / 2)
        draw.text((total_width / 2, y),
                  bottom_text,
                  fill=text_color,
                  font=font,
                  anchor="ms",
                  stroke_width=1 if best <= 14 else 0,
                  stroke_fill=stroke_color)

    if not os.path.exists(CONFIG["output_path"]):
        os.makedirs(CONFIG["output_path"])
    filepath = os.path.join(CONFIG["output_path"], f"{output}.png")
    image.save(filepath)
    get_logger().info(f"Image saved to: {filepath}")

    with open(filepath, "rb") as f:  # 'rb' 表示二进制读取
        image_bytes = f.read()  # 直接获取字节数据
    return image_bytes



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 19:57
# @Author  : Wu_RH
# @FileName: 3P.py
"""
[3](暂定):作者曾经在主群透露过的内容 雷指向的方向存在n个雷(不包括自己)
"""
from typing import Dict

from ....abs.Mrule import AbstractMinesClueRule, AbstractMinesValue
from ....abs.board import AbstractPosition, AbstractBoard
from ....utils.image_create import get_dummy, get_image, get_col, get_row
from ....utils.image_create import get_text as _get_text
from ....utils.tool import get_random


def get_text(
     text: str,
     width: float = "auto",
     height: float = "auto",
     cover_pos_label: bool = True,
     color: tuple[str, str] = ("#FFFF00", "#FF7F00"),
     dominant_by_height: bool = True
):
    return _get_text(
        text=text,
        width=width,
        height=height,
        cover_pos_label=cover_pos_label,
        color=color,
        dominant_by_height=dominant_by_height
    )


class Rule3P(AbstractMinesClueRule):
    name = ["3"]
    doc = "雷指向的方向存在n个雷(不包括自己)"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        for pos, _ in board("F"):
            choise = int(random.random() * 4)
            line = []
            match choise:
                case 0:
                    line = board.get_row_pos(pos)
                    line = line[1 + line.index(pos):]
                case 1:
                    line = board.get_row_pos(pos)
                    line = line[:line.index(pos)]
                case 2:
                    line = board.get_col_pos(pos)
                    line = line[1 + line.index(pos):]
                case 3:
                    line = board.get_col_pos(pos)
                    line = line[:line.index(pos)]

            value = board.batch(line, mode="type").count("F")
            obj = Value3P(pos, bytes([value | choise << 6]))
            board.set_value(pos, obj)
        return board


class Value3P(AbstractMinesValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        self.dir = code[0] >> 6
        self.pos = pos
        self.value = code[0] & 0x3f

    def __repr__(self):
        return "><V^"[self.dir] + str(self.value)

    def web_component(self, board: 'AbstractBoard') -> Dict:
        if self.dir in [0, 1]:
            return get_col(
                get_text("→←↓↑"[self.dir]),
                get_text(str(self.value))
            )
        else:
            return get_row(
                get_text("→←↓↑"[self.dir]),
                get_text(str(self.value))
            )

    def compose(self, board) -> Dict:
        match self.dir:
            case 3:
                # 上 ↑ ^
                if self.value == 1:
                    return get_row(
                            get_dummy(width=0.175),
                            get_text("1"),
                            get_image("up_flag"),
                            get_dummy(width=0.175),
                        )
                return get_row(
                        get_text(str(self.value)),
                        get_image("up_flag"),
                        spacing=-0.1,
                    )
            case 0:
                # 右 → >
                return get_col(
                    get_dummy(height=0.1),
                    get_image(
                        "right_flag",
                        image_height=0.2,
                        image_width=0.7
                    ),
                    get_dummy(height=-0.05),
                    get_text(str(self.value)),
                    get_dummy(height=0.2)
                )
            case 2:
                # 下 ↓ V
                if self.value == 1:
                    return get_row(
                            get_dummy(width=0.175),
                            get_text("1"),
                            get_image("down_flag"),
                            get_dummy(width=0.175),
                        )
                return get_row(
                        get_text(str(self.value)),
                        get_image("down_flag"),
                        spacing=-0.1,
                    )
            case 1:
                # 左 ← <
                return get_col(
                    get_dummy(height=0.1),
                    get_image(
                        "left_flag",
                        image_height=0.2,
                        image_width=0.7
                    ),
                    get_dummy(height=-0.05),
                    get_text(str(self.value)),
                    get_dummy(height=0.2)
                )
        return get_text("")

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        line = []
        match self.dir:
            case 0:
                line = board.get_row_pos(self.pos)
                line = line[1 + line.index(self.pos):]
            case 1:
                line = board.get_row_pos(self.pos)
                line = line[:line.index(self.pos)]
            case 2:
                line = board.get_col_pos(self.pos)
                line = line[1 + line.index(self.pos):]
            case 3:
                line = board.get_col_pos(self.pos)
                line = line[:line.index(self.pos)]
        var_list = board.batch(line, mode="variable")
        model.Add(sum(var_list) == self.value).OnlyEnforceIf(s)

    @classmethod
    def type(cls) -> bytes:
        return Rule3P.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value | self.dir << 6])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 18:32
# @Author  : Wu_RH
# @FileName: 2S.py
"""
[2S] 分段 (Segment)：每行有且仅有一组连续的雷
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule2S(AbstractMinesRule):
    name = ["2S", "分段", "Segment"]
    doc = "每行有且仅有一组连续的雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for key in board.get_interactive_keys():
            d_pos = board.boundary(key)
            for pos in board.get_col_pos(d_pos):
                line = board.get_row_pos(pos)
                line_var = board.batch(line, mode="variable")
                model.AddBoolOr(line_var)
                for index, var in enumerate(line_var[:-1]):
                    model.Add(sum(line_var[index + 1:]) == 0).OnlyEnforceIf([var, line_var[index + 1].Not(), s])
                for index, var in enumerate(line_var[1:]):
                    model.Add(sum(line_var[:index+1]) == 0).OnlyEnforceIf([var, line_var[index].Not(), s])



#!/usr/bin/env python3
"""
[3DD] 蛆 (Dual)：所有雷均形成1x2x1、2x1x1或1x1x2的三维矩形，即每个雷恰好有一个六连通相邻的雷
"""

from .. import Abstract3DMinesRule
from .....abs.board import AbstractBoard


class Rule3DD(Abstract3DMinesRule):  # type: ignore
    name = ["3DD", "3D1D", "蛆", "三维对偶"]
    doc = "所有雷均形成1x2x1、2x1x1或1x1x2的三维矩形"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="variable"):
            # 获取六连通邻域：同层四连通 + 上下层对应位置
            six_neighbors = self._get_six_connected_neighbors(board, pos)
            six_neighbor_vars = board.batch(six_neighbors, mode="variable", drop_none=True)

            if not six_neighbor_vars:
                # 如果没有六连通邻居，该位置不能是雷
                model.Add(var == 0)
                continue

            # 核心约束：每个雷恰好有1个六连通相邻的雷
            # val 为1时，六连通邻域中必须有且仅有一个雷
            sum_vals = sum(six_neighbor_vars)
            model.Add(sum_vals == 1).OnlyEnforceIf([var, s])

    def _get_six_connected_neighbors(self, board: 'AbstractBoard', pos):
        """获取六连通邻域：同层四连通 + 上下层对应位置"""
        neighbors = []

        # 同层四连通邻域（左右上下）
        four_neighbors = pos.neighbors(1)
        neighbors.extend(four_neighbors)

        # 上层对应位置
        up_pos = self.up(board, pos, n=1)
        if up_pos is not None:
            neighbors.append(up_pos)

        # 下层对应位置
        down_pos = self.down(board, pos, n=1)
        if down_pos is not None:
            neighbors.append(down_pos)

        return neighbors

    def suggest_total(self, info: dict):
        """建议雷的总数应该是偶数，因为雷成对出现"""
        def constraint_even_total(model, total):
            model.AddModuloEquality(0, total, 2)
        info["hard_fns"].append(constraint_even_total)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/20 14:12
# @Author  : Wu_RH
# @FileName: 2#.py
"""
[1#']包含以下规则:
[V], [1M], [1L], [1W], [1N], [1X], [1P], [1E]
"""
from minesweepervariants.abs.board import AbstractBoard
from . import AbstractClueSharp


class Rule1sharp(AbstractClueSharp):
    name = ["1#", "标签"]
    doc = "包含以下规则:[V], [1M], [1L], [1W], [1N], [1X], [1P], [1E]"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        rules_name = ["V", "1M", "1L", "1W", "1N", "1X", "1P", "1E"]
        super().__init__(rules_name, board, data)




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:21
# @Author  : Wu_RH
# @FileName: 1D'.py
"""
[1D'] 战舰 (Battleship)：每个雷区域为宽度为 1、长度不超过 4 的矩形，矩形不能对角相邻
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1Dp(AbstractMinesRule):
    name = ["1D'", "战舰", "Battleship"]
    doc = "每个雷区域为宽度为 1、长度不超过 4 的矩形，矩形不能对角相邻"

    def create_constraints(self, board: AbstractBoard, switch):
        # 获取求解器模型
        model = board.get_model()
        s = switch.get(model, self)

        for key in board.get_interactive_keys():
            boundary_pos = board.boundary(key)
            max_x = boundary_pos.x
            max_y = boundary_pos.y

            # 创建二维数组存储连续长度变量
            h_length = {}  # 水平连续长度
            v_length = {}  # 垂直连续长度

            # 遍历所有格子
            for y in range(max_y + 1):
                for x in range(max_x + 1):
                    pos = board.get_pos(x, y, key)
                    cell_var = board.get_variable(pos)

                    # === 对角相邻必须为非雷 ===
                    diagonals = []
                    # 左上
                    if x > 0 and y > 0:
                        diagonals.append(board.get_pos(x - 1, y - 1, key))
                    # 右上
                    if x < max_x and y > 0:
                        diagonals.append(board.get_pos(x + 1, y - 1, key))
                    # 左下
                    if x > 0 and y < max_y:
                        diagonals.append(board.get_pos(x - 1, y + 1, key))
                    # 右下
                    if x < max_x and y < max_y:
                        diagonals.append(board.get_pos(x + 1, y + 1, key))

                    for diag_pos in diagonals:
                        diag_var = board.get_variable(diag_pos)
                        model.Add(diag_var == 0).OnlyEnforceIf([cell_var, s])

                    # === 正交相邻雷数限制 (0-2) ===
                    ortho_vars = []
                    # 上
                    if y > 0:
                        ortho_vars.append(board.get_variable(board.get_pos(x, y - 1, key)))
                    # 下
                    if y < max_y:
                        ortho_vars.append(board.get_variable(board.get_pos(x, y + 1, key)))
                    # 左
                    if x > 0:
                        ortho_vars.append(board.get_variable(board.get_pos(x - 1, y, key)))
                    # 右
                    if x < max_x:
                        ortho_vars.append(board.get_variable(board.get_pos(x + 1, y, key)))

                    ortho_count = 0
                    if ortho_vars:
                        ortho_count = sum(ortho_vars)
                        model.Add(ortho_count <= 2).OnlyEnforceIf([cell_var, s])

                    # === 两个相邻雷必须同向 ===
                    # 创建方向指示变量
                    has_horizontal = model.NewBoolVar(f'hor_{x}_{y}')
                    has_vertical = model.NewBoolVar(f'ver_{x}_{y}')

                    # 水平相邻条件
                    if 0 < x < max_x:
                        left_var = board.get_variable(board.get_pos(x - 1, y, key))
                        right_var = board.get_variable(board.get_pos(x + 1, y, key))
                        model.Add(left_var + right_var == 2).OnlyEnforceIf([has_horizontal, s])
                        model.Add(has_horizontal == 0).OnlyEnforceIf([left_var.Not(), s])
                        model.Add(has_horizontal == 0).OnlyEnforceIf([right_var.Not(), s])
                    else:
                        has_horizontal = model.NewConstant(0)

                    # 垂直相邻条件
                    if 0 < y < max_y:
                        up_var = board.get_variable(board.get_pos(x, y - 1, key))
                        down_var = board.get_variable(board.get_pos(x, y + 1, key))
                        model.Add(up_var + down_var == 2).OnlyEnforceIf([has_vertical, s])
                        model.Add(has_vertical == 0).OnlyEnforceIf([up_var.Not(), s])
                        model.Add(has_vertical == 0).OnlyEnforceIf([down_var.Not(), s])
                    else:
                        has_vertical = model.NewConstant(0)

                    # 当有2个相邻雷时，必须是同一直线方向
                    if ortho_vars:
                        model.Add(ortho_count != 2).OnlyEnforceIf([has_horizontal.Not(), has_vertical.Not(), s])

                    # === 战舰长度限制 (1-4) ===
                    # 水平连续长度变量 (0-4)
                    h_len_var = model.NewIntVar(0, 4, f'hlen_{x}_{y}')
                    h_length[(x, y)] = h_len_var

                    # 水平长度约束
                    if x == 0:  # 最左列
                        # 当前是起点：长度=1(是雷)或0(非雷)
                        model.Add(h_len_var == cell_var).OnlyEnforceIf(s)
                    else:
                        left_pos = board.get_pos(x - 1, y, key)
                        left_var = board.get_variable(left_pos)

                        # 情况1：当前不是雷 → 长度为0
                        model.Add(h_len_var == 0).OnlyEnforceIf([cell_var.Not(), s])

                        # 情况2：当前是雷且左边不是雷 → 新起点，长度=1
                        model.Add(h_len_var == 1).OnlyEnforceIf([cell_var, left_var.Not(), s])

                        # 情况3：当前是雷且左边是雷 → 继承左边长度+1
                        if x > 0:
                            model.Add(h_len_var == h_length[(x - 1, y)] + 1).OnlyEnforceIf(
                                [cell_var, left_var, s])

                    # 垂直连续长度变量 (0-4)
                    v_len_var = model.NewIntVar(0, 4, f'vlen_{x}_{y}')
                    v_length[(x, y)] = v_len_var

                    # 垂直长度约束
                    if y == 0:  # 最上行
                        model.Add(v_len_var == cell_var).OnlyEnforceIf(s)
                    else:
                        up_pos = board.get_pos(x, y - 1, key)
                        up_var = board.get_variable(up_pos)

                        model.Add(v_len_var == 0).OnlyEnforceIf([cell_var.Not(), s])
                        model.Add(v_len_var == 1).OnlyEnforceIf([cell_var, up_var.Not(), s])
                        if y > 0:
                            model.Add(v_len_var == v_length[(x, y - 1)] + 1).OnlyEnforceIf(
                                [cell_var, up_var, s])

                    # === 长度上限约束 ===
                    # 当战舰达到最大长度(4)时，下一格必须不能是雷
                    # 水平方向
                    # if x < max_x - 1:  # 确保有下一格
                    #     next_right = board.get_pos(x + 1, y, key)
                    #     next_right_var = board.get_variable(next_right)
                    #     # 如果当前长度=4，下一格必须不是雷
                    #     model.Add(next_right_var == 0).OnlyEnforceIf(h_len_var == 4)
                    #
                    # # 垂直方向
                    # if y < max_y - 1:
                    #     next_down = board.get_pos(x, y + 1, key)
                    #     next_down_var = board.get_variable(next_down)
                    #     model.Add(next_down_var == 0).OnlyEnforceIf(v_len_var == 4)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/13 00:30
# @Author  : Wu_RH
# @FileName: -1M.py
"""
[*1M]: 雷分布将随机按照下述方式对称 [水平/垂直/对角/副对角/中心/旋转90度]对称
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rulen1M(AbstractMinesRule):
    name = ["*1M", "镜像", "Mirror"]
    doc = "雷分布将随机按照下述方式对称 [水平/垂直/对角/副对角/中心]对称"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        tmp_a = model.NewBoolVar("[*1M]")      # 垂直对称
        tmp_b = model.NewBoolVar("[*1M]")      # 水平对称
        tmp_c = model.NewBoolVar("[*1M]")      # 正斜角对称
        tmp_d = model.NewBoolVar("[*1M]")      # 副斜角对称
        tmp_e = model.NewBoolVar("[*1M]")      # 中心对称

        for key in board.get_interactive_keys():
            pos_bound = board.boundary(key)
            for index_x in range(pos_bound.x + 1):
                for index_y in range(pos_bound.y + 1):
                    var = board.get_variable(board.get_pos(index_x, index_y, key))
                    if var is None:
                        continue
                    var_a = board.get_variable(board.get_pos(index_x, pos_bound.y-index_y, key))
                    var_b = board.get_variable(board.get_pos(pos_bound.x-index_x, index_y, key))
                    var_c = board.get_variable(board.get_pos(pos_bound.y-index_y, pos_bound.x-index_x, key))
                    var_d = board.get_variable(board.get_pos(index_y, index_x, key))
                    var_e = board.get_variable(board.get_pos(pos_bound.x-index_x, pos_bound.y-index_y, key))
                    if var_a is not None: model.Add(var == var_a).OnlyEnforceIf(tmp_a)
                    if var_b is not None: model.Add(var == var_b).OnlyEnforceIf(tmp_b)
                    if var_c is not None: model.Add(var == var_c).OnlyEnforceIf(tmp_c)
                    if var_d is not None: model.Add(var == var_d).OnlyEnforceIf(tmp_d)
                    if var_e is not None: model.Add(var == var_e).OnlyEnforceIf(tmp_e)

        model.AddBoolOr([tmp_a, tmp_b, tmp_c, tmp_d, tmp_e]).OnlyEnforceIf(s)



from typing import Literal, Never, NotRequired, TypedDict, Optional, List, Tuple, Dict, Union

from flask.config import T

__all__ = ["CellType", "CellState", "Board", "CountInfo", "ComponentTemplate", "ComponentConfig", "CellConfig",
           "BoardMetadata", "U_Hint", "ClickResponse"]


class CellType(TypedDict):
    boardname: str
    x: int
    y: int


class CellState(TypedDict):
    type: CellType
    isRevealed: bool
    isLoading: bool
    hint1: bool
    hint2: bool
    error: bool
    errormine: bool


class Board(TypedDict):
    name: Optional[str]
    position: Tuple[int, int]
    showLabel: Optional[bool]
    showName: Optional[bool]
    dye: Optional[List[List[bool]]]
    mask: Optional[List[List[bool]]]
    size: Tuple[int, int]


class CountInfo(TypedDict):
    total: int
    known: Optional[int]
    unknown: int
    remains: Optional[int]


class ComponentTemplate(TypedDict):
    name: str
    value: object


class ComponentConfig(TypedDict):
    type: Literal["container", "text", "assets", "template"]
    value: Union[List['ComponentConfig'], str, ComponentTemplate]
    style: Optional[str]
    class_: Optional[str]


class CellConfig(TypedDict):
    overlayText: str
    position: CellType
    component: ComponentConfig
    highlight: Optional[Dict[str, List[CellType]]]


class BoardMetadata(TypedDict):
    rules: List[str]
    boards: Dict[str, Board]
    cells: List[CellConfig]
    count: Optional[CountInfo]
    seed: Optional[str]
    noFail: Optional[bool]
    noHint: Optional[bool]
    mode: Literal["NORMAL", "EXPERT", "ULTIMATE", "PUZZLE", "UNKNOWN"]
    u_mode: NotRequired[List[str]]
    u_hint: NotRequired[Dict[str, int]]
    version: tuple[int, int, int]


class U_Hint(TypedDict):
    emptycount: int
    flagcount: int
    markcount: NotRequired[int]


class ClickResponse(TypedDict):
    success: bool
    gameover: bool
    reason: str
    cells: List[CellConfig]
    count: Optional[CountInfo]
    noFail: Optional[bool]
    noHint: Optional[bool]
    mines: NotRequired[List[CellType]]
    win: NotRequired[bool]
    u_hint: NotRequired[U_Hint]


class GenerateBoardResult(TypedDict):
    reason: str
    success: bool


class CreateGameParams(TypedDict):
    size: str
    rules: str
    mode: str
    total: str
    u_mode: NotRequired[str]
    dye: NotRequired[str]
    mask: NotRequired[str]
    seed: NotRequired[str]


class ClickData(TypedDict):
    x: int
    y: int
    boardName: str
    button: Literal["left", "right", "middle"]


type MetadataResult = BoardMetadata

type ResponseType[T] = T | tuple[T, int]



from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.game import Mode, UMode, ValueAsterisk, MinesAsterisk
from minesweepervariants.server._typing import BoardMetadata
from minesweepervariants.utils.impl_obj import VALUE_QUESS, MINES_TAG


def init_component(data: dict, *, color: str = '--primary-color', invalid: bool = False) -> dict:
    if data["type"] in ["col", "row"]:
        style = "display: flex; "
        if data["type"] == "col":
            # 垂直布局：主轴垂直方向，交叉轴居中
            style += "flex-direction: column;"
            # 子项高度平均分配（使用 flex-grow 实现）
            for child in data["children"]:
                child_style = init_component(child, color=color, invalid=invalid).get("style", "")
                if "flex-grow" not in child_style:
                    child_style += " flex-grow: 1;"
                child["style"] = child_style
        else:
            # 水平布局：主轴水平方向，交叉轴居中
            style += "flex-direction: row;"
            # 子项宽度平均分配（使用 flex-grow 实现）
            for child in data["children"]:
                child_style = init_component(child, color=color, invalid=invalid).get("style", "")
                if "flex-grow" not in child_style:
                    child_style += " flex-grow: 1;"
                child["style"] = child_style
        style += " align-items: center; justify-content: center; gap: 5%;"
        style += " width: 100%; height: 100%; flex-grow: 1;"
        return {
            "type": "container",
            "value": [init_component(i, color=color, invalid=invalid) for i in data["children"]],
            "style": style
        }

    elif data["type"] == "text":
        # 文本项：使用 flex 布局填充可用空间，添加溢出处理
        style = (f"color: rgb(from var({color}) r g b / "
                 f"{50 if invalid else 100}%); text-align: center;")
        style += " display: flex; justify-content: center; align-items: center;"
        style += " flex: 1; min-width: 0; max-width: 100%;"  # 关键：允许内容收缩
        style += " overflow: hidden; text-overflow: ellipsis; white-space: nowrap;"
        style += data["style"]

        return {
            "type": "text",
            "value": data.get("text", ""),
            "style": style
        }

    elif data["type"] == "image":
        path = data.get("image")
        # 图片项：保持比例，居中显示
        return {
            "type": "assets",
            "value": path,
            "style": f"fill: rgb(from var({color}) r g b / "
                     f"{50 if invalid else 100}%); flex: 1; min-width: 0;"
                     f"{data['style']}"
        }

    elif data["type"] == "placeholder":
        style = "flex-shrink: 0;"  # 防止占位符被压缩

        if "width" in data:
            # 行内容器中使用固定宽度
            style += f" width: {int(100 * data['width'])}%;"
        if "height" in data:
            # 列内容器中使用固定高度
            style += f" height: {int(100 * data['height'])}%;"

        return {
            "type": "container",
            "value": [],
            "style": style
        }
    elif data["type"] == "template":
        return data
    raise ValueError("Unknown component type")


def format_cell(_board, pos, label):
    obj = _board[pos]
    color = "--flag-color" if _board.get_type(pos) == "F" else "--foreground-color"
    invalid = False if obj is None else obj.invalid(_board)
    cell_data = init_component({
        "type": "row",
        "children": [obj.web_component(_board)]
    }, color=color, invalid=invalid)
    cell_data["style"] += " width: 100%; height: 100%; align-items: center; justify-content: center;"
    cell_data["style"] += (f"color: rgb(from var({color}) r g b /"
                           f" {50 if invalid else 100}%);"
                           f" flex: 1; min-width: 0;")
    # if dye:
    #     cell_data["style"] += " background-color: rgb(from var(--foreground-color) r g b / 29%);"
    if (
        _board.get_config(pos.board_key, "pos_label") and
        (
            _board.get_type(pos) == "F" or
            _board.get_type(pos) == "N"
        )
    ):
        txt = chr(64 + pos.y // 26) if pos.y > 25 else ''
        txt += chr(65 + pos.y % 26)
        txt += f"={pos.x}"
        cell_data = {
            "type": "container",
            "value": [
                cell_data,
                {
                    "type": "template",
                    "style": "",
                    "value": {
                        "name": "backgroundStr",
                        "value": txt
                    },
                }
            ],
            "style": ""
        }
    VALUE = _board.get_config(pos.board_key, "VALUE")
    MINES = _board.get_config(pos.board_key, "MINES")
    if obj in [VALUE, MINES, None]:
        overlayText = ""
    else:
        overlayText = obj.type().decode("ascii")
    # hightlight = [{
    #             "x": pos.x,
    #             "y": pos.y,
    #             "boardname": pos.board_key,
    #         }]
    hightlight = {pos.board_key: [[pos.x, pos.y]]}
    if obj is not None:
        if obj.high_light(_board) is not None:
            for h_pos in set(h_pos for h_pos in obj.high_light(_board) if _board.in_bounds(h_pos)):
                # hightlight.append({
                #     "x": h_pos.x,
                #     "y": h_pos.y,
                #     "boardname": h_pos.board_key,
                # })
                if h_pos.board_key not in hightlight:
                    hightlight[h_pos.board_key] = []
                hightlight[h_pos.board_key].append([h_pos.x, h_pos.y])
    cell_data = {
        "type": "" if obj is None else obj.type().decode("ascii"),
        "position": {
            "x": pos.x, "y": pos.y,
            "boardname": pos.board_key
        },
        "component": cell_data,
        "highlight": hightlight,
        "clickable": True,
        "overlayText": overlayText if label else ""
    }
    # import json
    # json_str = json.dumps(cell_data, separators=(",", ":"))
    # print(json_str)
    return cell_data


def format_board(_board: AbstractBoard):
    if _board is None:
        return

    boards = {}
    cells = []
    count = 0
    for key in _board.get_board_keys():
        dye_list = [
            [_board.get_dyed(pos) if _board.is_valid(pos) else False
             for pos in _board.get_row_pos(col_pos)]
            for col_pos in _board.get_col_pos(
                _board.boundary(key=key)
            )
        ]
        mask_list = [
            [not _board.is_valid(pos) for pos in _board.get_row_pos(col_pos)]
            for col_pos in _board.get_col_pos(
                _board.boundary(key=key)
            )
        ]
        boards[key] = {
            "size": _board.get_config(key, "size"),
            "position": [_board.get_board_keys().index(key), 0],
            "showLabel": _board.get_config(key, "row_col"),
            "showName": not _board.get_config(key, "row_col"),  # TODO 何时不显示Name?
            # TODO X=N, poslabel
        }
        print(mask_list)
        if any(any(i) for i in mask_list):
            boards[key].update({
                "mask": mask_list
            })
        if any(any(i) for i in dye_list):
            boards[key].update({
                "dye": dye_list,
            })
        for pos, obj in _board(key=key):
            if obj is None:
                if _board.get_config(key, "pos_label"):
                    txt = chr(64 + pos.y // 26) if pos.y > 25 else ''
                    txt += chr(65 + pos.y % 26)
                    txt += f"={pos.x}"
                    cells.append({
                        "type": "",
                        "position": {
                            "x": pos.x, "y": pos.y,
                            "boardname": pos.board_key
                        },
                        "component": {
                            "type": "template",
                            "style": "",
                            "value": {
                                "name": "backgroundStr",
                                "value": txt
                            },
                        },
                        "highlight": {},
                        "clickable": True,
                        "overlayText": ""
                    })
                continue
            label = obj not in [
                VALUE_QUESS, MINES_TAG,
                _board.get_config(key, "MINES"),
                _board.get_config(key, "VALUE"),
            ]
            label = (
                    _board.get_config(key, "by_mini") and
                    label and
                    not (
                            isinstance(obj, ValueAsterisk) or
                            isinstance(obj, MinesAsterisk)
                    )
            )
            cells.append(
                format_cell(_board, pos, label))
            count += 1
    return boards, cells, count


def format_gamemode(gamemode: Mode, u_gamemode: UMode):
    u_mode: list[str] = []

    match gamemode:
        case Mode.NORMAL:
            mode = "NORMAL"
        case Mode.EXPERT:
            mode = "EXPERT"
        case Mode.ULTIMATE:
            mode = "ULTIMATE"
            if u_gamemode & UMode.ULTIMATE_A:
                u_mode.append("+A")
            if u_gamemode & UMode.ULTIMATE_F:
                u_mode.append("+F")
            if u_gamemode & UMode.ULTIMATE_S:
                u_mode.append("+S")
            if u_gamemode & UMode.ULTIMATE_R:
                u_mode.append("+R")
            if u_gamemode & UMode.ULTIMATE_P:
                u_mode.append("+!")
        case Mode.PUZZLE:
            mode = "PUZZLE"
        case _:
            mode = "UNKNOWN"
    return mode, u_mode



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/15 11:59
# @Author  : Wu_RH
# @FileName: 1T''.py

"""
[1T'']纯三连：每个雷属于唯一的三连组(横/竖/斜)
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1Tpp(AbstractMinesRule):
    name = ["1T''", "纯三连"]
    doc = "每个雷属于唯一的三连组(横/竖/斜)"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        # 实现: 每个雷格必须且只能属于一个三连 (横/竖/对角)
        position_coverage = {pos: [] for pos, _ in board()}

        # 生成所有可能的三连 (8 个方向)
        for pos, _ in board():
            directions = [
                (0, 1), (0, -1), (1, 0), (-1, 0),  # 横/竖
                (1, 1), (1, -1), (-1, 1), (-1, -1)  # 对角
            ]

            for dx, dy in directions:
                positions = [
                    pos,
                    pos.shift(dx, dy),
                    pos.shift(2 * dx, 2 * dy)
                ]

                if all(board.is_valid(p) for p in positions):
                    var_list = [board.get_variable(p) for p in positions]

                    # 三连变量
                    b = model.NewBoolVar(f"triple_{pos}_{dx}_{dy}")
                    model.AddBoolAnd(var_list).OnlyEnforceIf([b, s])

                    for p in positions:
                        position_coverage[p].append(b)

        # 对每个位置，覆盖它的三连数量必须等于该位置是否为雷 (0 或 1)
        for pos, var in board(mode="variable"):
            coverage_list = position_coverage[pos]
            if coverage_list:
                # 将 Bool vars 的和与位置变量相等
                model.Add(sum(coverage_list) == var).OnlyEnforceIf(s)
            else:
                # 若位置没有任何三连覆盖，则不能为雷
                model.Add(var == 0).OnlyEnforceIf(s)

    def suggest_total(self, info: dict):

        def hard_constraint(m, total):
            m.AddModuloEquality(0, total, 3)

        info["hard_fns"].append(hard_constraint)



import time
import asyncio
from typing import Any
from orjson import loads, dumps

class DataStore:
    def __init__(self, path: str):
        import asyncio
        self.path = path
        self.data = {}
        self.file = open(path, "w+b")

        self._readers = 0
        self._readers_lock = asyncio.Lock()
        self._writer_lock = asyncio.Lock()
        self.last_save = 0
        self.modified = False

    async def get(self, key: str) -> str | None:
        async with self._readers_lock:
            self._readers += 1
            if self._readers == 1:
                await self._writer_lock.acquire()
        try:
            return self.data.get(key)
        finally:
            async with self._readers_lock:
                self._readers -= 1
                if self._readers == 0:
                    self._writer_lock.release()

    async def set(self, key: str, value: Any) -> None:
        async with self._writer_lock:
            self.data[key] = value
            self.modified = True

    async def load(self):
        self.file.seek(0)
        content = self.file.read()
        if content:
            async with self._writer_lock:
                self.data = loads(content)
                self.modified = False

    async def save(self, force=False):
        t = time.time()
        if force or (self.modified and time.time() - self.last_save > 10):
            print("Auto saving...")
            self.last_save = t
            self.file.seek(0)
            self.file.truncate()
            self.file.write(dumps(self.data))
            self.file.flush()
            self.modified = False

    async def close(self):
        await self._writer_lock.acquire()
        self.file.close()

    async def init(self):
        await self.load()
        self.last_save = time.time()

    async def start(self):
        while self.file.closed is False:
            await self.save()
            await asyncio.sleep(10)

    def __del__(self):
        self.file.close()
        print("DataStore closed")


"""
[3L] 环：所有雷构成一条环。环是一条宽度为 1 的八连通路径，不存在分叉和交叉, 环的头尾相连
"""
from ....abs.Lrule import AbstractMinesRule

from .connect import connect


class Rule1S(AbstractMinesRule):
    name = ["3L", "环", "Loop"]
    doc = "所有雷构成一条环。环是一条宽度为 1 的八连通路径，不存在分叉和交叉，环的头尾相连"

    def create_constraints(self, board, switch):
        model = board.get_model()
        s = switch.get(model, self)

        connect(
            model=model,
            board=board,
            connect_value=1,
            nei_value=2,
            switch=s,
        )

        for pos, var in board(mode="variable"):
            var_list = board.batch(pos.neighbors(2), mode="variable", drop_none=True)
            model.Add(sum(var_list) == 2).OnlyEnforceIf([var, s])



from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition

from .connect import connect

class Rule1C(AbstractMinesRule):
    name = ["1C'", "C'", "八连通'", "Connected'"]
    doc = "(1) 染色格的雷和 (2) 非染色格的雷各自八连通。染色区域和非染色区域至少有一个雷。"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s1 = switch.get(model, self)
        s2 = switch.get(model, self)
        dyed_vars = []
        undyed_vars = []
        for pos, var in board(mode="variable"):
            dyed = board.get_dyed(pos)
            if dyed:
                dyed_vars.append((pos, var))
                undyed_vars.append((pos, model.NewConstant(0)))
            else:
                dyed_vars.append((pos, model.NewConstant(0)))
                undyed_vars.append((pos, var))
        model.AddBoolOr([d[1] for d in dyed_vars]).OnlyEnforceIf(s1)
        model.AddBoolOr([u[1] for u in undyed_vars]).OnlyEnforceIf(s2)
        connect(model, board, s1, positions_vars=dyed_vars)
        connect(model, board, s2, positions_vars=undyed_vars)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/18 14:01
# @Author  : Wu_RH
# @FileName: 3I.py
"""
[3Y]阴阳(Yin-Yang):所有雷四连通，所有非雷四连通，不存在2*2的雷或非雷
"""
from typing import List

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition
from .connect import connect


def block(a_pos: AbstractPosition, board: AbstractBoard) -> List[AbstractPosition]:
    b_pos = a_pos.up()
    c_pos = a_pos.left()
    d_pos = b_pos.left()
    if not board.in_bounds(d_pos):
        return []
    return [a_pos, b_pos, c_pos, d_pos]


class Rule3Y(AbstractMinesRule):
    name = ["3Y", "阴阳", "Yin-Yang"]
    doc = "所有雷四连通，所有非雷四连通，不存在2*2的雷或非雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        positions_vars = [(pos, var) for pos, var in board("always", mode="variable")]

        connect(
            model,
            board,
            ub=len(positions_vars),
            connect_value=0,
            nei_value=1,
            switch=s
        )
        connect(
            model,
            board,
            ub=len(positions_vars),
            connect_value=1,
            nei_value=1,
            switch=s
        )

        # 大定式
        for pos, _ in board():
            pos_list = block(pos, board)
            if not pos_list:
                continue
            a, b, c, d = board.batch(pos_list, mode="variable")
            model.AddBoolOr([a.Not(), b, c, d.Not()]).OnlyEnforceIf(s)  # 排除 1010
            model.AddBoolOr([a, b.Not(), c.Not(), d]).OnlyEnforceIf(s)  # 排除 0101
            model.AddBoolOr([a, b, c, d]).OnlyEnforceIf(s)  # 排除 0000
            model.AddBoolOr([a.Not(), b.Not(), c.Not(), d.Not()]).OnlyEnforceIf(s)  # 排除 1111

    def suggest_total(self, info: dict):
        ub = 0
        for key in info["interactive"]:
            size = info["size"][key]
            ub += size[0] * size[1]
        info["soft_fn"](ub * 0.5, 0)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:37
# @Author  : Wu_RH
# @FileName: 2H.py
"""
[2H] 横向 (Horizontal)：所有雷必须存在横向相邻的雷
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule2H(AbstractMinesRule):
    name = ["2H", "横向", "Horizontal"]
    doc = "所有雷必须存在横向相邻的雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="variable"):
            if not board.in_bounds(pos.right()):
                model.Add(board.get_variable(pos.left()) == 1).OnlyEnforceIf([var, s])
            elif not board.in_bounds(pos.left()):
                model.Add(board.get_variable(pos.right()) == 1).OnlyEnforceIf([var, s])
            else:
                model.AddBoolOr(board.batch([pos.right(), pos.left()], mode="variable")).OnlyEnforceIf([var, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 11:40
# @Author  : xxx
# @FileName: sharp.py
"""
[#]标签: 线索会有不同的规则, 每个线索的规则会单独标出
"""
from typing import List, Tuple, Optional

from ....abs.Mrule import AbstractMinesClueRule
from ....abs.board import AbstractBoard
from ....abs.rule import AbstractRule
from ....utils.tool import get_random


class RuleSharp(AbstractMinesClueRule):
    name = ["#", "标签", "Tag"]
    doc = "线索会有不同的规则, 每个线索的规则会单独标出"

    def __init__(self, board: AbstractBoard, data: list[AbstractMinesClueRule]):
        super().__init__(board, None)
        self.rules = data
        for key in board.get_interactive_keys():
            board.set_config(key, "by_mini", True)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        boards = []
        for rule in self.rules:
            boards.append(rule.fill(board.clone()))
        for key in board.get_board_keys():
            for pos, _ in board("F", key=key):
                values = [_board.get_value(pos)
                          for _board in boards
                          if _board.get_type(pos) != "N"]
                if not values:
                    continue
                board.set_value(pos, get_random().choice(values))
        return board

    def combine(self, rules: List[Tuple['AbstractRule', Optional[str]]]):
        for rule in self.rules:
            rule.combine(rules)

    def suggest_total(self, info: dict):
        for rule in self.rules:
            rule.suggest_total(info)

    def init_board(self, board: 'AbstractBoard'):
        for rule in self.rules:
            rule.init_board(board)

    def init_clear(self, board: 'AbstractBoard'):
        for rule in self.rules:
            rule.init_clear(board)

    def create_constraints(self, board: 'AbstractBoard', switch):
        for rule in self.rules:
            rule.create_constraints(board, switch)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/11 21:51
# @Author  : Wu_RH
# @FileName: 3E.py

"""
[3E]演化: 每个3x1区域决定其中间下方1格是否为雷，演化被当前题板所有区域共享。
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG

NAME_3E = ["3E_x", "3E_y"]


class Rule3E(AbstractMinesRule):
    name = ["3E", "迭代", "演化"]
    doc = "每个3x1区域决定其中间下方1格是否为雷，演化被当前题板所有区域共享。"

    def __init__(self, board: AbstractBoard, data=None):
        super().__init__(board, data)
        board.generate_board(NAME_3E[0], size=(8, 3))
        board.generate_board(NAME_3E[1], size=(8, 1))
        board.set_config(NAME_3E[1], "VALUE", VALUE_QUESS)
        board.set_config(NAME_3E[1], "MINES", MINES_TAG)

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()

        vals_list = []

        for i in range(8):
            x = model.NewIntVar(i, i, f"[3E]{i}_x")
            y = board.get_variable(board.get_pos(i, 0, NAME_3E[1]))
            s_pos = board.get_pos(i, 1, NAME_3E[0])
            model.Add(x == i)
            s = switch.get(model, s_pos)
            vals_list.append([x, y, s])

        for key in board.get_interactive_keys():
            col = board.get_col_pos(board.boundary(key=key))
            bool_var_index = 0
            for pos in col:
                row = board.get_row_pos(pos)
                for index in range(len(row) - 2):
                    variables = board.batch(row[index: index + 3], mode="variable")
                    y_variables = board.get_variable(row[index + 1].down())
                    if y_variables is None:
                        continue
                    for vals in vals_list:
                        var_x, var_y, s = vals
                        cond = model.NewBoolVar(f"[3E]{bool_var_index}")
                        model.Add(
                            var_x == 4 * variables[0] + 2 * variables[1] + variables[2]
                        ).OnlyEnforceIf([cond, s])
                        model.Add(
                            var_x != 4 * variables[0] + 2 * variables[1] + variables[2]
                        ).OnlyEnforceIf([cond.Not(), s])
                        model.Add(y_variables == var_y).OnlyEnforceIf([cond, s])
                        bool_var_index += 1

    def init_board(self, board: 'AbstractBoard'):
        for index, pos in enumerate(board.get_col_pos(
                board.get_pos(0, 0, NAME_3E[0]))):
            pos1, pos2, pos3 = board.get_row_pos(pos)
            board.set_value(pos1, MINES_TAG if index & 4 else VALUE_QUESS)
            board.set_value(pos2, MINES_TAG if index & 2 else VALUE_QUESS)
            board.set_value(pos3, MINES_TAG if index & 1 else VALUE_QUESS)
        y_col = board.get_col_pos(board.boundary(key=NAME_3E[1]))
        for key in board.get_interactive_keys():
            col = board.get_col_pos(board.boundary(key=key))
            for pos in col:
                row = board.get_row_pos(pos)
                for index in range(len(row) - 2):
                    types = board.batch(row[index: index + 3], mode="type")
                    y_type = board.get_type(row[index + 1].down())
                    if y_type == "":
                        continue
                    index = 4 if types[0] == "F" else 0
                    index += 2 if types[1] == "F" else 0
                    index += 1 if types[2] == "F" else 0
                    board.set_value(y_col[index], MINES_TAG if y_type == "F" else VALUE_QUESS)

    def init_clear(self, board: 'AbstractBoard'):
        for pos, _ in board(key=NAME_3E[1]):
            board.set_value(pos, None)



"""
[3M]对映：线索表示周围八格组成四个正好相对的组合中有且仅有一个是雷的数量（题板外视为非雷）
"""

from ....abs.board import AbstractBoard, AbstractPosition
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....utils.tool import get_logger
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG

offsets = [
    (-1, -1),
    (-1, 0),
    (-1, 1),
    (0, -1),
    (0, 1),
    (1, -1),
    (1, 0),
    (1, 1)
]


class Rule3M(AbstractClueRule):
    name = ["3M", "对映", "Mirror"]
    doc = "线索表示周围八格组成四个正好相对的格子组有且仅有一个是雷的组数（题板外视为非雷）"

    logger = get_logger()

    def fill(self, board: AbstractBoard) -> AbstractBoard:
        logger = get_logger()
        for pos, _ in board("N"):
            temp = []
            for x, y in offsets:
                temp.append(board.get_type(pos.shift(x, y)) == "F")
            value = sum(1 for i in range(4) if temp[i] != temp[7 - i])
            board.set_value(pos, Value3M(pos, count=value))
            logger.debug(f"Set value for {pos}: 3M[{value}]")
        return board

class Value3M(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count
        self.neighbor = self.pos.neighbors(2)

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule3M.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def invalid(self, board: 'AbstractBoard') -> bool:
        return board.batch(self.neighbor, mode="type").count("N") == 0

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        types = []
        for x, y in offsets:
            type_obj = board.get_type(self.pos.shift(x, y))
            if type_obj == "C" or type_obj == "":
                types.append(-1)
            elif type_obj == "F":
                types.append(1)
            else:
                types.append(2)
        defined = 0
        can_be_deduced = []
        for i in range(4):
            if types[i] + types[7 - i] == 0:
                defined += 1
            elif types[i] == 2 and types[7 - i] != 2:
                can_be_deduced.append(i)
            elif types[i] != 2 and types[7 - i] == 2:
                can_be_deduced.append(7 - i)
        if defined == self.count:
            for i in can_be_deduced:
                offsetX, offsetY = offsets[i]
                mirrorType = board.get_type(self.pos.shift(-offsetX, -offsetY))
                if mirrorType == "C" or mirrorType == "":
                    board.set_value(self.pos.shift(offsetX, offsetY), VALUE_QUESS)
                elif mirrorType == "F":
                    board.set_value(self.pos.shift(offsetX, offsetY), MINES_TAG)
            return True
        elif defined + len(can_be_deduced) == self.count:
            for i in can_be_deduced:
                offsetX, offsetY = offsets[i]
                mirrorType = board.get_type(self.pos.shift(-offsetX, -offsetY))
                if mirrorType == "C" or mirrorType == "":
                    board.set_value(self.pos.shift(offsetX, offsetY), MINES_TAG)
                elif mirrorType == "F":
                    board.set_value(self.pos.shift(offsetX, offsetY), VALUE_QUESS)
            return True
        return False

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()

        g1 = model.NewBoolVar("3M_g1")
        g2 = model.NewBoolVar("3M_g2")
        g3 = model.NewBoolVar("3M_g3")
        g4 = model.NewBoolVar("3M_g4")

        s = switch.get(model, self)

        model.Add(sum(board.batch([self.pos.shift(-1, -1), self.pos.shift(1, 1)], mode="variable",
                                  drop_none=True)) == 1).OnlyEnforceIf([g1, s])
        model.Add(sum(board.batch([self.pos.shift(-1, -1), self.pos.shift(1, 1)], mode="variable",
                                  drop_none=True)) != 1).OnlyEnforceIf([g1.Not(), s])

        model.Add(sum(board.batch([self.pos.shift(-1, 0), self.pos.shift(1, 0)], mode="variable",
                                  drop_none=True)) == 1).OnlyEnforceIf([g2, s])
        model.Add(sum(board.batch([self.pos.shift(-1, 0), self.pos.shift(1, 0)], mode="variable",
                                  drop_none=True)) != 1).OnlyEnforceIf([g2.Not(), s])

        model.Add(sum(board.batch([self.pos.shift(-1, 1), self.pos.shift(1, -1)], mode="variable",
                                  drop_none=True)) == 1).OnlyEnforceIf([g3, s])
        model.Add(sum(board.batch([self.pos.shift(-1, 1), self.pos.shift(1, -1)], mode="variable",
                                  drop_none=True)) != 1).OnlyEnforceIf([g3.Not(), s])

        model.Add(sum(board.batch([self.pos.shift(0, -1), self.pos.shift(0, 1)], mode="variable",
                                  drop_none=True)) == 1).OnlyEnforceIf([g4, s])
        model.Add(sum(board.batch([self.pos.shift(0, -1), self.pos.shift(0, 1)], mode="variable",
                                  drop_none=True)) != 1).OnlyEnforceIf([g4.Not(), s])

        model.Add(sum([g1, g2, g3, g4]) == self.count).OnlyEnforceIf(s)
